# This file is auto-generated by scripts/generate_sync_client.py
# Type stub for the synchronous REST client.

from __future__ import annotations

from typing import Any, Dict, List, Optional, Tuple, Type, Union
from pydantic import BaseModel
from decimal import Decimal
from ethereal.models.config import RESTConfig
from ethereal.chain_client import ChainClient
from uuid import UUID
from ethereal.models.rest import (
    AccountSignerQuotaDto,
    BalanceHistoryDto,
    CancelOrderDto,
    CancelOrderResultDto,
    DryRunOrderCreatedDto,
    FundingDto,
    InitiateWithdrawDto,
    LinkSignerDto,
    MarketLiquidityDto,
    MarketPriceDto,
    OrderDto,
    OrderFillDto,
    PositionDto,
    PositionFundingHistoryDto,
    PositionLiquidationDto,
    ProductDto,
    ProjectedFundingDto,
    RevokeLinkedSignerDto,
    RpcConfigDto,
    SignerDto,
    SubaccountBalanceDto,
    SubaccountDto,
    SubmitOrderCreatedDto,
    SubmitOrderDto,
    TokenDto,
    TradeDto,
    TransferDto,
    UnrealizedPnlHistoryDto,
    VolumeHistoryDto,
    WithdrawDto,
)

class RESTClient:
    def __init__(self, config: Union[Dict[str, Any], RESTConfig] = ...) -> None: ...
    def close(self) -> None: ...
    def __enter__(self) -> RESTClient: ...
    def __exit__(self, exc_type, exc_val, exc_tb) -> None: ...
    def cancel_all_orders(
        self, subaccount_id: UUID, product_ids: Optional[List[UUID]] = None, **kwargs
    ) -> List[CancelOrderResultDto]: ...
    def cancel_order(
        self, order_to_cancel: CancelOrderDto, **kwargs
    ) -> List[CancelOrderResultDto]: ...
    def cancel_orders(
        self,
        order_ids: List[UUID],
        sender: Optional[str] = None,
        subaccount: Optional[str] = None,
        client_order_ids: List[str] = [],
        sign: bool = True,
        submit: bool = True,
        **kwargs,
    ) -> Union[List[CancelOrderResultDto], CancelOrderDto]: ...
    def create_order(
        self,
        order_type: str,
        quantity: float,
        side: int,
        price: Optional[float] = None,
        ticker: Optional[str] = None,
        product_id: Optional[UUID] = None,
        client_order_id: Optional[str] = None,
        sender: Optional[str] = None,
        subaccount: Optional[str] = None,
        time_in_force: Optional[str] = None,
        post_only: Optional[bool] = None,
        reduce_only: Optional[bool] = False,
        close: Optional[bool] = None,
        stop_price: Optional[float] = None,
        stop_type: Optional[int] = None,
        expires_at: Optional[int] = None,
        group_id: Optional[str] = None,
        group_contingency_type: Optional[int] = None,
        sign: bool = True,
        dry_run: bool = False,
        submit: bool = True,
    ) -> Union[SubmitOrderCreatedDto, DryRunOrderCreatedDto, SubmitOrderDto]: ...
    def delete(
        self,
        url_path: Any,
        params: Union[dict, None] = None,
        data: Union[dict, None] = None,
        base_url_override: Union[str, None] = None,
        **kwargs,
    ) -> Dict[str, Any]: ...
    def dry_run_order(
        self, order: SubmitOrderDto, **kwargs
    ) -> DryRunOrderCreatedDto: ...
    def get(
        self,
        url_path: Any,
        params: Union[dict, None] = None,
        base_url_override: Union[str, None] = None,
        **kwargs,
    ) -> Dict[str, Any]: ...
    def get_maintenance_margin(
        self,
        subaccount_id: UUID,
        positions: Optional[List[PositionDto] | List[Dict[str, Any]]] = None,
        products: Optional[List[ProductDto] | List[Dict[str, Any]]] = None,
        product_ids: Optional[List[UUID]] = None,
    ) -> Decimal: ...
    def get_market_liquidity(self, **kwargs) -> MarketLiquidityDto: ...
    def get_order(self, id: UUID, **kwargs) -> OrderDto: ...
    def get_position(self, id: UUID, **kwargs) -> PositionDto: ...
    def get_projected_funding(self, **kwargs) -> ProjectedFundingDto: ...
    def get_rpc_config(self, **kwargs) -> RpcConfigDto: ...
    def get_signer(self, id: UUID, **kwargs) -> SignerDto: ...
    def get_signer_quota(self, **kwargs) -> AccountSignerQuotaDto: ...
    def get_subaccount(self, id: UUID, **kwargs) -> SubaccountDto: ...
    def get_subaccount_balance_history(self, **kwargs) -> List[BalanceHistoryDto]: ...
    def get_subaccount_balances(self, **kwargs) -> List[SubaccountBalanceDto]: ...
    def get_subaccount_funding_history(
        self, **kwargs
    ) -> List[PositionFundingHistoryDto]: ...
    def get_subaccount_unrealized_pnl_history(
        self, **kwargs
    ) -> List[UnrealizedPnlHistoryDto]: ...
    def get_subaccount_volume_history(self, **kwargs) -> List[VolumeHistoryDto]: ...
    def get_token(self, id: UUID, **kwargs) -> TokenDto: ...
    def get_tokens(self) -> List[TokenDto]: ...
    def get_validated(
        self,
        url_path: Any,
        request_model: Type[BaseModel],
        response_model: Type[BaseModel],
        base_url_override: Union[str, None] = None,
        **kwargs,
    ) -> BaseModel: ...
    def link_linked_signer(self, dto: LinkSignerDto, **kwargs) -> SignerDto: ...
    def list_fills(self, **kwargs) -> List[OrderFillDto]: ...
    def list_funding(self, **kwargs) -> List[FundingDto]: ...
    def list_market_prices(self, **kwargs) -> List[MarketPriceDto]: ...
    def list_orders(self, **kwargs) -> List[OrderDto]: ...
    def list_position_liquidations(self, **kwargs) -> List[PositionLiquidationDto]: ...
    def list_positions(self, **kwargs) -> List[PositionDto]: ...
    def list_products(self, **kwargs) -> List[ProductDto]: ...
    def list_signers(self, **kwargs) -> List[SignerDto]: ...
    def list_subaccounts(self, **kwargs) -> List[SubaccountDto]: ...
    def list_token_transfers(self, **kwargs) -> List[TransferDto]: ...
    def list_token_withdraws(self, **kwargs) -> List[WithdrawDto]: ...
    def list_tokens(self, **kwargs) -> List[TokenDto]: ...
    def list_trades(self, **kwargs) -> List[TradeDto]: ...
    def post(
        self,
        url_path: Any,
        params: Union[dict, None] = None,
        data: Union[dict, None] = None,
        base_url_override: Union[str, None] = None,
        **kwargs,
    ) -> Dict[str, Any]: ...
    def prepare_and_send_request(
        self,
        http_method: Any,
        url_path: Any,
        params: Union[dict, None] = None,
        data: Union[dict, None] = None,
        base_url_override: Union[str, None] = None,
    ) -> Any: ...
    def prepare_cancel_order(
        self,
        sender: str,
        subaccount: str,
        order_ids: List[UUID] = [],
        client_order_ids: List[str] = [],
        include_signature: bool = False,
        **kwargs,
    ) -> CancelOrderDto: ...
    def prepare_linked_signer(
        self,
        sender: str,
        signer: str,
        subaccount: str,
        subaccount_id: UUID,
        signer_signature: str = "",
        include_signature: bool = False,
        **kwargs,
    ) -> LinkSignerDto: ...
    def prepare_order(
        self,
        sender: str,
        price: Union[str, float, Decimal, None] = None,
        quantity: Union[str, float, Decimal, None] = None,
        side: Union[int, None] = None,
        subaccount: Union[str, None] = None,
        onchain_id: Union[float, None] = None,
        order_type: Union[str, None] = None,
        client_order_id: Union[str, None] = None,
        time_in_force: Union[str, None] = None,
        post_only: Union[bool, None] = False,
        reduce_only: Union[bool, None] = False,
        close: Union[bool, None] = None,
        stop_price: Union[str, float, Decimal, None] = None,
        stop_type: Union[int, None] = None,
        group_id: Union[str, None] = None,
        group_contingency_type: Union[int, None] = None,
        expires_at: Union[int, None] = None,
        include_signature: bool = False,
        **kwargs,
    ) -> SubmitOrderDto: ...
    def prepare_revoke_linked_signer(
        self,
        sender: str,
        signer: str,
        subaccount: str,
        subaccount_id: UUID,
        include_signature: bool = False,
        **kwargs,
    ) -> RevokeLinkedSignerDto: ...
    def prepare_withdraw_token(
        self,
        subaccount: str,
        token: str,
        amount: int,
        account: str,
        destination_address: str,
        destination_endpoint: int,
        include_signature: bool = False,
        nonce: Union[str, None] = None,
        signed_at: Union[int, None] = None,
    ) -> InitiateWithdrawDto: ...
    def products(self, refresh: bool = False) -> List[ProductDto]: ...
    def products_by_id(self, refresh: bool = False) -> Dict[UUID, ProductDto]: ...
    def products_by_ticker(self, refresh: bool = False) -> Dict[str, ProductDto]: ...
    def put(
        self,
        url_path: Any,
        params: Union[dict, None] = None,
        data: Union[dict, None] = None,
        base_url_override: Union[str, None] = None,
        **kwargs,
    ) -> Dict[str, Any]: ...
    def replace_order(
        self,
        order: Optional[OrderDto] = None,
        order_id: Optional[UUID] = None,
        quantity: Optional[float] = None,
        price: Optional[float] = None,
        time_in_force: Optional[str] = None,
        post_only: Optional[bool] = None,
        reduce_only: Optional[bool] = False,
    ) -> Tuple[SubmitOrderCreatedDto, bool]: ...
    def revoke_linked_signer(
        self, dto: RevokeLinkedSignerDto, **kwargs
    ) -> SignerDto: ...
    def send_request(
        self,
        http_method: Any,
        url_path: Any,
        params: Any,
        headers: Any,
        data: Any = None,
        base_url_override: Union[str, None] = None,
    ) -> Any: ...
    def sign_cancel_order(
        self, order_to_cancel: CancelOrderDto, private_key: Union[str, None] = None
    ) -> CancelOrderDto: ...
    def sign_linked_signer(
        self,
        link_to_sign: LinkSignerDto,
        signer_private_key: Union[str, None] = None,
        private_key: Union[str, None] = None,
    ) -> LinkSignerDto: ...
    def sign_order(
        self, order: SubmitOrderDto, private_key: Union[str, None] = None
    ) -> SubmitOrderDto: ...
    def sign_revoke_linked_signer(
        self, revoke_to_sign: RevokeLinkedSignerDto
    ) -> RevokeLinkedSignerDto: ...
    def sign_withdraw_token(
        self, withdraw_dto: InitiateWithdrawDto, private_key: Union[str, None] = None
    ) -> InitiateWithdrawDto: ...
    def subaccounts(self, refresh: bool = False) -> List[SubaccountDto]: ...
    def submit_order(
        self, order: SubmitOrderDto, **kwargs
    ) -> SubmitOrderCreatedDto: ...
    def tokens(self, refresh: bool = False) -> List[TokenDto]: ...
    def withdraw_token(
        self, dto: InitiateWithdrawDto, token_id: UUID, **kwargs
    ) -> WithdrawDto: ...
    def _get_pages(
        self,
        endpoint: str,
        request_model: Type[BaseModel],
        response_model: Type[BaseModel],
        paginate: bool = False,
        **kwargs,
    ) -> Any: ...

    # Exposed internals
    @property
    def chain(self) -> Optional[ChainClient]: ...
    @property
    def logger(self) -> Any: ...
    @property
    def provider(self) -> Any: ...
    @property
    def private_key(self) -> Optional[str]: ...
    @property
    def _models(self) -> Any: ...



from typing import List
from uuid import UUID

from ethereal.constants import API_PREFIX
from ethereal.models.rest import PositionDto, PositionLiquidationDto


async def list_positions(
    self,
    **kwargs,
) -> List[PositionDto]:
    """Lists positions for a subaccount.

    Args:
        subaccount_id (str): UUID of the subaccount. Required.

    Other Parameters:
        product_ids (List[str], optional): UUIDs of products to filter by. Optional.
        open (bool, optional): Filter for open positions. Optional.
        created_after (float, optional): Filter positions created after this timestamp. Optional.
        created_before (float, optional): Filter positions created before this timestamp. Optional.
        side (int, optional): Filter by position side (0 for buy, 1 for sell). Optional.
        is_liquidated (bool, optional): Filter liquidated positions. Optional.
        order (str, optional): Sort order, 'asc' or 'desc'. Optional.
        limit (float, optional): Maximum number of results to return. Optional.
        cursor (str, optional): Pagination cursor for fetching the next page. Optional.
        order_by (str, optional): Field to order by, e.g., 'size', 'createdAt', 'updatedAt', or 'realizedPnl'. Optional.
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        List[PositionDto]: List of position information for the subaccount.
    """
    res = await self.get_validated(
        url_path=f"{API_PREFIX}/position",
        request_model=self._models.V1PositionGetParametersQuery,
        response_model=self._models.PageOfPositionDtos,
        **kwargs,
    )
    data = [
        self._models.PositionDto(**model.model_dump(by_alias=True))
        for model in res.data
    ]
    return data


async def get_position(
    self,
    id: UUID,
    **kwargs,
) -> PositionDto:
    """Gets a specific position by ID.

    Args:
        id (str): UUID of the position. Required.

    Other Parameters:
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        PositionDto: Position information for the specified ID.
    """
    endpoint = f"{API_PREFIX}/position/{id}"
    res = await self.get(endpoint, **kwargs)
    return self._models.PositionDto(**res)


async def list_position_liquidations(
    self,
    **kwargs,
) -> List[PositionLiquidationDto]:
    """Lists position liquidations.

    Other Parameters:
        order (str, optional): Sort order, 'asc' or 'desc'. Optional.
        limit (float, optional): Maximum number of results to return. Optional.
        cursor (str, optional): Pagination cursor for fetching the next page. Optional.
        order_by (str, optional): Field to order by (e.g., 'createdAt'). Optional.
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        List[PositionLiquidationDto]: List of position liquidation records.
    """
    res = await self.get_validated(
        url_path=f"{API_PREFIX}/position/liquidation",
        request_model=self._models.V1PositionLiquidationGetParametersQuery,
        response_model=self._models.PageOfPositionLiquidationsDto,
        **kwargs,
    )
    data = [
        self._models.PositionLiquidationDto(**model.model_dump(by_alias=True))
        for model in res.data
    ]
    return data



import pytest
from web3 import Web3


def test_provider(rc):
    rc.logger.info(f"Chain ID: {rc.chain.chain_id}")
    assert rc.provider is not None
    assert isinstance(rc.provider, Web3)


def test_block(rc):
    """Test block method."""
    block = rc.provider.eth.get_block("latest")
    assert block is not None
    assert block.get("number") is not None
    assert block.get("hash") is not None


def test_nonce(rc):
    """Test nonce method."""
    nonce = rc.chain.get_nonce(rc.chain.address)
    rc.logger.info(f"Nonce: {nonce}")
    assert nonce is not None


def test_gas(rc):
    """Test gas methods."""
    gas_price = rc.provider.eth.gas_price
    rc.logger.info(f"Gas Price: {gas_price}")
    assert gas_price is not None
    assert gas_price > 0

    max_priority_fee = rc.provider.eth.max_priority_fee
    rc.logger.info(f"Max Priority Fee: {max_priority_fee}")
    assert max_priority_fee is not None

    gas_limit = rc.provider.eth.estimate_gas(
        {"from": rc.chain.address, "to": rc.chain.address, "value": 1}
    )
    rc.logger.info(f"Gas Limit: {gas_limit}")
    assert gas_limit is not None
    assert gas_limit > 0


def test_eth_balance(rc):
    """Test eth balance method."""
    balance = rc.chain.get_balance(rc.chain.address)
    rc.logger.info(f"Balance: {balance}")
    assert balance is not None
    assert balance >= 0


def test_usde_balance(rc):
    """Test token balance method."""
    balance = rc.chain.get_token_balance(rc.chain.address, rc.chain.usde.address)
    rc.logger.info(f"Balance: {balance}")
    assert balance is not None
    assert balance >= 0


def test_deposit_usde(rc):
    """Test USDe deposit."""
    deposit_tx = rc.chain.deposit_usde(100)
    rc.logger.info(f"Deposit Tx: {deposit_tx}")

    assert deposit_tx is not None
    assert deposit_tx.get("data") is not None
    assert deposit_tx.get("value") == rc.chain.provider.to_wei(100, "ether")
    assert rc.provider.is_checksum_address(deposit_tx.get("from"))
    assert rc.provider.is_checksum_address(deposit_tx.get("to"))


def test_deposit_usde_with_account_name_bytes(rc):
    """Test USDe deposit with account_name_bytes."""
    account_name_bytes = rc.chain.provider.to_hex(text="primary").ljust(66, "0")
    deposit_tx = rc.chain.deposit_usde(100, account_name_bytes=account_name_bytes)
    rc.logger.info(f"Deposit Tx: {deposit_tx}")

    assert deposit_tx is not None
    assert deposit_tx.get("data") is not None
    assert deposit_tx.get("value") == rc.chain.provider.to_wei(100, "ether")
    assert rc.provider.is_checksum_address(deposit_tx.get("from"))
    assert rc.provider.is_checksum_address(deposit_tx.get("to"))


def test_deposit_usde_with_both_params_fails(rc):
    """Test USDe deposit fails when both account_name and account_name_bytes are provided."""
    account_name_bytes = rc.chain.provider.to_hex(text="custom").ljust(66, "0")

    with pytest.raises(
        ValueError, match="Cannot provide both account_name and account_name_bytes"
    ):
        rc.chain.deposit_usde(
            100, account_name="custom", account_name_bytes=account_name_bytes
        )



from typing import Optional
from pydantic import BaseModel, Field, AnyHttpUrl, AnyWebsocketUrl, field_validator


class BaseConfig(BaseModel):
    """Base client configuration model."""

    verbose: bool = False
    network: str = Field(
        default="default",
        description="Network to connect to (testnet, devnet, mainnet)",
    )

    @field_validator("network")
    @classmethod
    def validate_network(cls, v: str) -> str:
        """Validate network is supported."""
        valid_networks = ["default", "testnet", "devnet", "mainnet", "local"]
        if v not in valid_networks:
            raise ValueError(f"Network must be one of {valid_networks}, got '{v}'")
        return v


class HTTPConfig(BaseConfig):
    """HTTP client configuration model."""

    base_url: Optional[AnyHttpUrl] = Field(
        default=None,
        description="Base API URL (auto-set based on network if not provided)",
    )
    timeout: int = Field(default=30, gt=0, description="Request timeout in seconds")
    rate_limit_headers: bool = Field(
        default=False, description="Include rate limit headers in responses"
    )


class WSBaseConfig(BaseConfig):
    """Base Websocket client configuration model."""

    base_url: AnyWebsocketUrl = Field(
        default=AnyWebsocketUrl("wss://ws.etherealtest.net/"), description="Base WS URL"
    )


class ChainConfig(BaseConfig):
    """Blockchain configuration model."""

    rpc_url: AnyHttpUrl = Field(default=..., description="RPC endpoint URL")
    address: Optional[str] = Field(
        default=None, description="Blockchain address for transactions"
    )
    private_key: Optional[str] = Field(
        default=None, description="Private key for blockchain transactions"
    )


class RESTConfig(HTTPConfig):
    """REST client configuration model."""

    chain_config: Optional[ChainConfig] = Field(
        default=None, description="Blockchain client configuration"
    )
    default_time_in_force: str = Field(
        default="GTD", description="Default time in force for orders"
    )
    default_post_only: bool = Field(
        default=False, description="Default post-only flag for orders"
    )
    archive_base_url: Optional[AnyHttpUrl] = Field(
        default=None,
        description="Base URL for archive REST requests (auto-set if not provided)",
    )


class WSConfig(WSBaseConfig):
    """Websocket client configuration model."""

    pass



import pytest
import asyncio


def test_ws_open(ws):
    """Test opening a websocket connection."""
    assert ws.sio is not None


@pytest.mark.asyncio
async def test_ws_emit(ws):
    """Test emitting an event to the server."""

    def log_callback(data):
        ws.logger.info(f"Received data: {data}")

    ws.subscribe(
        stream_type="BookDepth",
        product_id="4368f737-eeed-40bc-bb1b-7b9bfd88a057",
        callback=log_callback,
    )

    ws.subscribe(
        stream_type="OrderFill",
        product_id="4368f737-eeed-40bc-bb1b-7b9bfd88a057",
        subaccount_id="f051803c-a388-4ae5-815c-ad69338ead86",
        callback=log_callback,
    )

    # await until a log appears
    await asyncio.sleep(2)
    assert ws.sio is not None



import pytest
from ethereal.rest.util import is_uuid, client_order_id_to_bytes32


def test_is_uuid_valid_with_dashes():
    """Test valid UUID string with dashes."""
    uuid_str = "550e8400-e29b-41d4-a716-446655440000"
    assert is_uuid(uuid_str) is True


def test_is_uuid_valid_without_dashes():
    """Test valid UUID string without dashes."""
    uuid_str = "550e8400e29b41d4a716446655440000"
    assert is_uuid(uuid_str) is False


def test_is_uuid_invalid_too_short():
    """Test invalid UUID - too short."""
    uuid_str = "550e8400-e29b-41d4"
    assert is_uuid(uuid_str) is False


def test_is_uuid_invalid_too_long():
    """Test invalid UUID - too long."""
    uuid_str = "550e8400-e29b-41d4-a716-446655440000-extra"
    assert is_uuid(uuid_str) is False


def test_is_uuid_invalid_bad_format():
    """Test invalid UUID - bad format."""
    uuid_str = "not-a-uuid-at-all"
    assert is_uuid(uuid_str) is False


def test_is_uuid_invalid_wrong_characters():
    """Test invalid UUID - wrong characters."""
    uuid_str = "550g8400-e29b-41d4-a716-446655440000"
    assert is_uuid(uuid_str) is False


def test_client_order_id_none_input_raises_error():
    """Test None input raises ValueError."""
    with pytest.raises(ValueError, match="Client order ID cannot be None"):
        client_order_id_to_bytes32(None)


def test_client_order_id_short_string_conversion():
    """Test short string gets padded to 32 bytes."""
    test_str = "hello"
    result = client_order_id_to_bytes32(test_str)

    # Should be 0x + 64 hex chars (32 bytes)
    assert result.startswith("0x")
    assert len(result) == 66

    # Decode the hex to verify content
    hex_part = result[2:]  # Remove 0x prefix
    decoded_bytes = bytes.fromhex(hex_part)

    # Should start with "hello" and be padded with null bytes
    assert decoded_bytes.startswith(b"hello")
    assert decoded_bytes.rstrip(b"\0") == b"hello"
    assert len(decoded_bytes) == 32


def test_client_order_id_max_length_string():
    """Test 32-character string conversion."""
    test_str = (
        "order_id_with_exactly_32_chars12"  # Exactly 32 characters, clearly not UUID
    )
    result = client_order_id_to_bytes32(test_str)

    assert result.startswith("0x")
    assert len(result) == 66

    # Decode and verify
    hex_part = result[2:]
    decoded_bytes = bytes.fromhex(hex_part)
    assert decoded_bytes == test_str.encode("utf-8")


def test_client_order_id_string_too_long_raises_error():
    """Test string longer than 32 characters raises ValueError."""
    test_str = "a" * 33  # 33 characters - too long

    with pytest.raises(
        ValueError, match="Client order ID cannot be longer than 32 characters, got 33"
    ):
        client_order_id_to_bytes32(test_str)


def test_client_order_id_empty_string_raises_error():
    """Test empty string raises ValueError."""
    with pytest.raises(ValueError, match="Client order ID cannot be empty"):
        client_order_id_to_bytes32("")


def test_client_order_id_alphanumeric_string():
    """Test alphanumeric string conversion."""
    test_str = "order123"
    result = client_order_id_to_bytes32(test_str)

    assert result.startswith("0x")
    assert len(result) == 66

    hex_part = result[2:]
    decoded_bytes = bytes.fromhex(hex_part)
    assert decoded_bytes.startswith(test_str.encode("utf-8"))
    assert decoded_bytes.rstrip(b"\0") == test_str.encode("utf-8")


def test_client_order_id_unicode_string():
    """Test Unicode string conversion."""
    test_str = "hello🚀"  # Contains emoji
    result = client_order_id_to_bytes32(test_str)

    assert result.startswith("0x")
    assert len(result) == 66

    hex_part = result[2:]
    decoded_bytes = bytes.fromhex(hex_part)

    # UTF-8 encoding of emoji takes multiple bytes
    encoded = test_str.encode("utf-8")
    assert len(encoded) <= 32  # Should fit in 32 bytes
    assert decoded_bytes.startswith(encoded)


def test_client_order_id_string_too_long_non_uuid():
    """Test that non-UUID long strings raise error."""
    fake_uuid = "not-a-real-uuid-string-here-fake-and-too-long"  # 43 characters

    # This should be treated as a string (too long, should raise error)
    with pytest.raises(ValueError):
        client_order_id_to_bytes32(fake_uuid)


def test_client_order_id_uuid_formats_same_result():
    """Test edge cases around UUID detection."""
    # Valid UUID without dashes
    uuid_no_dash = "550e8400e29b41d4a716446655440000"
    result1 = client_order_id_to_bytes32(uuid_no_dash)
    assert result1.startswith("0x")

    # Same as UUID with dashes
    uuid_with_dash = "550e8400-e29b-41d4-a716-446655440000"
    result2 = client_order_id_to_bytes32(uuid_with_dash)
    assert result2.startswith("0x")

    # The results are different
    assert result1 != result2



"""Pure async REST API tests for submitting and managing orders."""

import asyncio
import time
from decimal import Decimal
from typing import List
import uuid
import pytest


def safe_round_price(price, tick_size):
    """Round price to valid tick size using decimal precision."""
    from decimal import Decimal as D, ROUND_HALF_UP

    if tick_size == 0:
        return float(price)

    price_decimal = D(str(price))
    tick_size_decimal = D(str(tick_size))
    return float(
        (price_decimal / tick_size_decimal).quantize(D("1"), rounding=ROUND_HALF_UP)
        * tick_size_decimal
    )


@pytest.mark.asyncio
async def test_rest_limit_order_floats_submit_cancel(
    async_rc, async_subaccount, network
):
    subaccount = async_subaccount
    pid = (await async_rc.list_products())[0].id
    tick_size = float((await async_rc.products_by_id())[pid].tick_size)
    prices = (await async_rc.list_market_prices(product_ids=[pid]))[0]
    best_bid_price = float(prices.best_bid_price)
    bid_price = safe_round_price(best_bid_price * 0.90, tick_size)
    order = await async_rc.create_order(
        order_type="LIMIT",
        product_id=pid,
        side=0,
        price=bid_price,
        quantity=0.003,
        time_in_force="GTD",
        post_only=False,
        sender=async_rc.chain.address,
        subaccount=subaccount.name,
    )
    assert isinstance(order, async_rc._models.SubmitOrderCreatedDto)
    cancelled = await async_rc.cancel_orders(
        sender=async_rc.chain.address, subaccount=subaccount.name, order_ids=[order.id]
    )
    assert isinstance(cancelled, List)
    assert all(isinstance(o, async_rc._models.CancelOrderResultDto) for o in cancelled)


@pytest.mark.asyncio
async def test_rest_limit_order_decimal_submit_cancel(
    async_rc, async_subaccount, network
):
    subaccount = async_subaccount
    products = await async_rc.list_products()
    pid = next(p.id for p in products if getattr(p, "ticker", None) == "ETHUSD")
    tick_size = float((await async_rc.products_by_id())[pid].tick_size)
    prices = (await async_rc.list_market_prices(product_ids=[pid]))[0]
    best_bid_price = float(prices.best_bid_price)
    bid_price = safe_round_price(best_bid_price * 0.90, tick_size)
    expires_at = int(time.time()) + 3600
    order = await async_rc.create_order(
        order_type="LIMIT",
        product_id=pid,
        side=0,
        price=Decimal(str(bid_price)),
        quantity=Decimal("0.003"),
        time_in_force="GTD",
        expires_at=expires_at,
        post_only=False,
        sender=async_rc.chain.address,
        subaccount=subaccount.name,
    )
    assert isinstance(order, async_rc._models.SubmitOrderCreatedDto)
    fetched = await async_rc.get_order(order.id)
    assert isinstance(fetched, async_rc._models.OrderDto)
    assert fetched.id == order.id and int(fetched.expires_at) == expires_at
    cancelled = await async_rc.cancel_orders(
        sender=async_rc.chain.address, subaccount=subaccount.name, order_ids=[order.id]
    )
    assert isinstance(cancelled, List)
    assert all(isinstance(o, async_rc._models.CancelOrderResultDto) for o in cancelled)


@pytest.mark.asyncio
async def test_async_limit_order_with_client_order_id_uuid(async_rc, async_subaccount):
    """Create an order with a clientOrderId, fetch it, and verify filtering by client id."""
    client = async_rc
    sub = async_subaccount

    # Resolve product
    products = await client.list_products()
    pid = products[0].id

    # Compute a valid limit price
    products_by_id = {p.id: p for p in products}
    tick_size = float(products_by_id[pid].tick_size)
    prices = (await client.list_market_prices(product_ids=[pid]))[0]
    best_bid = float(prices.best_bid_price)
    price = float(Decimal(str(best_bid)) * Decimal("0.90"))
    if tick_size:
        # Round to tick size
        q = Decimal(str(price)) / Decimal(str(tick_size))
        price = float((q.quantize(Decimal("1"))) * Decimal(str(tick_size)))

    # Unique client order id
    coid = str(uuid.uuid4())

    # Create order with client_order_id
    created = await client.create_order(
        order_type="LIMIT",
        product_id=pid,
        side=0,
        price=Decimal(str(price)),
        quantity=Decimal("0.003"),
        time_in_force="GTD",
        post_only=False,
        client_order_id=coid,
        sender=client.chain.address,
        subaccount=sub.name,
    )

    assert created.client_order_id == coid

    # Verify we can fetch the order by id
    fetched = await client.get_order(created.id)
    assert fetched.id == created.id

    # Verify filtering by client order id returns the created order
    orders = await client.list_orders(subaccount_id=sub.id, client_order_id=coid)
    assert any(o.id == created.id for o in orders)

    # Cancel
    await client.cancel_orders(
        sender=client.chain.address,
        subaccount=sub.name,
        order_ids=[],
        client_order_ids=[coid],
    )

    # Verify the order is canceled
    fetched_cancel = await client.get_order(created.id)
    assert fetched_cancel.status.value == "CANCELED"


@pytest.mark.asyncio
async def test_async_limit_order_with_client_order_id_string(
    async_rc, async_subaccount
):
    """Create an order with a clientOrderId, fetch it, and verify filtering by client id specified as a non-UUID string."""
    client = async_rc
    sub = async_subaccount

    # Resolve product
    products = await client.list_products()
    pid = products[0].id

    # Compute a valid limit price
    products_by_id = {p.id: p for p in products}
    tick_size = float(products_by_id[pid].tick_size)
    prices = (await client.list_market_prices(product_ids=[pid]))[0]
    best_bid = float(prices.best_bid_price)
    price = float(Decimal(str(best_bid)) * Decimal("0.90"))
    if tick_size:
        # Round to tick size
        q = Decimal(str(price)) / Decimal(str(tick_size))
        price = float((q.quantize(Decimal("1"))) * Decimal(str(tick_size)))

    # Unique client order id
    coid = str(uuid.uuid4()).replace("-", "")

    # Create order with client_order_id
    created = await client.create_order(
        order_type="LIMIT",
        product_id=pid,
        side=0,
        price=Decimal(str(price)),
        quantity=Decimal("0.003"),
        time_in_force="GTD",
        post_only=False,
        client_order_id=coid,
        sender=client.chain.address,
        subaccount=sub.name,
    )

    assert created.client_order_id == coid

    # Verify we can fetch the order by id
    fetched = await client.get_order(created.id)
    assert fetched.id == created.id

    # Verify filtering by client order id returns the created order
    orders = await client.list_orders(subaccount_id=sub.id, client_order_id=coid)
    assert any(o.id == created.id for o in orders)

    # Cancel
    await client.cancel_orders(
        sender=client.chain.address,
        subaccount=sub.name,
        order_ids=[],
        client_order_ids=[coid],
    )

    # Verify the order is canceled
    fetched_cancel = await client.get_order(created.id)
    assert fetched_cancel.status.value == "CANCELED"


@pytest.mark.asyncio
async def test_async_limit_order_with_client_order_id_string_short(
    async_rc, async_subaccount
):
    """Create an order with a clientOrderId, fetch it, and verify filtering by client id specified as a non-UUID string under 32 characters."""
    client = async_rc
    sub = async_subaccount

    # Resolve product
    products = await client.list_products()
    pid = products[0].id

    # Compute a valid limit price
    products_by_id = {p.id: p for p in products}
    tick_size = float(products_by_id[pid].tick_size)
    prices = (await client.list_market_prices(product_ids=[pid]))[0]
    best_bid = float(prices.best_bid_price)
    price = float(Decimal(str(best_bid)) * Decimal("0.90"))
    if tick_size:
        # Round to tick size
        q = Decimal(str(price)) / Decimal(str(tick_size))
        price = float((q.quantize(Decimal("1"))) * Decimal(str(tick_size)))

    # Unique client order id
    coid = str(uuid.uuid4()).replace("-", "")[:20] + "ZZZZ"

    # Create order with client_order_id
    created = await client.create_order(
        order_type="LIMIT",
        product_id=pid,
        side=0,
        price=Decimal(str(price)),
        quantity=Decimal("0.003"),
        time_in_force="GTD",
        post_only=False,
        client_order_id=coid,
        sender=client.chain.address,
        subaccount=sub.name,
    )

    assert created.client_order_id == coid

    # Verify we can fetch the order by id
    fetched = await client.get_order(created.id)
    assert fetched.id == created.id

    # Verify filtering by client order id returns the created order
    orders = await client.list_orders(subaccount_id=sub.id, client_order_id=coid)
    assert any(o.id == created.id for o in orders)

    # Cancel
    await client.cancel_orders(
        sender=client.chain.address,
        subaccount=sub.name,
        order_ids=[],
        client_order_ids=[coid],
    )

    # Verify the order is canceled
    fetched_cancel = await client.get_order(created.id)
    assert fetched_cancel.status.value == "CANCELED"


@pytest.mark.asyncio
async def test_rest_limit_order_submit_cancel_multiple(
    async_rc, async_subaccount, network
):
    subaccount = async_subaccount
    pid = (await async_rc.list_products())[0].id
    tick_size = float((await async_rc.products_by_id())[pid].tick_size)
    prices = (await async_rc.list_market_prices(product_ids=[pid]))[0]
    best_bid_price = float(prices.best_bid_price)
    bid_price = safe_round_price(best_bid_price * 0.90, tick_size)
    ids = []
    for _ in range(2):
        o = await async_rc.create_order(
            order_type="LIMIT",
            product_id=pid,
            side=0,
            price=bid_price,
            quantity=0.003,
            time_in_force="GTD",
            post_only=False,
            sender=async_rc.chain.address,
            subaccount=subaccount.name,
        )
        assert isinstance(o, async_rc._models.SubmitOrderCreatedDto)
        ids.append(o.id)
    cancelled = await async_rc.cancel_orders(
        sender=async_rc.chain.address, subaccount=subaccount.name, order_ids=ids
    )
    assert isinstance(cancelled, List)
    assert all(isinstance(o, async_rc._models.CancelOrderResultDto) for o in cancelled)


@pytest.mark.asyncio
async def test_rest_limit_order_submit_cancel_all(async_rc, async_subaccount, network):
    sub = async_subaccount
    pid = (await async_rc.list_products())[0].id
    try:
        await async_rc.cancel_all_orders(
            sender=async_rc.chain.address, subaccount_id=sub.id, product_ids=[pid]
        )
    except Exception:
        pass
    tick_size = float((await async_rc.products_by_id())[pid].tick_size)
    prices = (await async_rc.list_market_prices(product_ids=[pid]))[0]
    best = float(prices.best_bid_price)
    ids = []
    for i in range(5):
        price = safe_round_price(best * (0.90 + i * 0.01), tick_size)
        o = await async_rc.create_order(
            order_type="LIMIT",
            product_id=pid,
            side=0,
            price=price,
            quantity=0.003,
            sender=async_rc.chain.address,
            subaccount=sub.name,
        )
        ids.append(o.id)
    cancelled = await async_rc.cancel_all_orders(
        sender=async_rc.chain.address, subaccount_id=sub.id, product_ids=[pid]
    )
    assert isinstance(cancelled, List)
    assert len(cancelled) == len(ids)
    await asyncio.sleep(1)
    for oid in ids:
        ord = await async_rc.get_order(id=oid)
        assert ord.status.value == "CANCELED"


@pytest.mark.asyncio
async def test_rest_limit_order_submit_cancel_all_specify_products(async_rc, network):
    subaccounts = await async_rc.list_subaccounts(sender=async_rc.chain.address)
    sub = subaccounts[0]
    products = await async_rc.list_products()
    order_ids = {}
    for i in range(2):
        pid = products[i].id
        info = (await async_rc.products_by_id())[pid]
        tick = Decimal(str(info.tick_size))
        min_quantity = Decimal(str(info.min_quantity))
        prices = (await async_rc.list_market_prices(product_ids=[pid]))[0]
        best = Decimal(str(prices.best_bid_price))
        order_ids[pid] = []
        for j in range(5):
            price = safe_round_price(best * Decimal(str(0.90 + j * 0.01)), tick)
            o = await async_rc.create_order(
                order_type="LIMIT",
                product_id=pid,
                side=0,
                price=price,
                quantity=min_quantity,
                time_in_force="GTD",
                post_only=False,
                sender=async_rc.chain.address,
                subaccount=sub.name,
            )
            order_ids[pid].append(o.id)
    cancelled = await async_rc.cancel_all_orders(
        sender=async_rc.chain.address,
        subaccount_id=sub.id,
        product_ids=[products[0].id],
    )
    assert len(cancelled) == len(order_ids[products[0].id])
    await asyncio.sleep(1)
    for oid in order_ids[products[0].id]:
        ord = await async_rc.get_order(id=oid)
        assert ord.status.value == "CANCELED"
    for oid in order_ids[products[1].id]:
        ord = await async_rc.get_order(id=oid)
        assert ord.status.value == "NEW"
    cancelled2 = await async_rc.cancel_all_orders(
        sender=async_rc.chain.address,
        subaccount_id=sub.id,
        product_ids=[products[1].id],
    )
    assert len(cancelled2) == len(order_ids[products[1].id])
    await asyncio.sleep(1)
    for oid in order_ids[products[1].id]:
        ord = await async_rc.get_order(id=oid)
        assert ord.status.value == "CANCELED"


@pytest.mark.asyncio
async def test_rest_limit_order_submit_cancel_all_multiple_products(async_rc, network):
    subaccounts = await async_rc.list_subaccounts(sender=async_rc.chain.address)
    sub = subaccounts[0]
    products = await async_rc.list_products()
    ids = []
    for i in range(2):
        pid = products[i].id
        info = (await async_rc.products_by_id())[pid]
        tick = Decimal(str(info.tick_size))
        min_quantity = Decimal(str(info.min_quantity))
        prices = (await async_rc.list_market_prices(product_ids=[pid]))[0]
        best = Decimal(str(prices.best_bid_price))
        for j in range(5):
            price = safe_round_price(best * Decimal(str(0.90 + j * 0.01)), tick)
            o = await async_rc.create_order(
                order_type="LIMIT",
                product_id=pid,
                side=0,
                price=price,
                quantity=min_quantity,
                time_in_force="GTD",
                post_only=False,
                sender=async_rc.chain.address,
                subaccount=sub.name,
            )
            ids.append(o.id)
    cancelled = await async_rc.cancel_all_orders(
        sender=async_rc.chain.address, subaccount_id=sub.id
    )
    assert isinstance(cancelled, List) and len(cancelled) == len(ids)
    await asyncio.sleep(1)
    for oid in ids:
        ord = await async_rc.get_order(id=oid)
        assert ord.status.value == "CANCELED"


@pytest.mark.asyncio
async def test_rest_limit_order_dry(async_rc, network):
    subaccounts = await async_rc.subaccounts()
    pid = (await async_rc.list_products())[0].id
    tick_size = float((await async_rc.products_by_id())[pid].tick_size)
    prices = (await async_rc.list_market_prices(product_ids=[pid]))[0]
    best = float(prices.best_bid_price)
    price = safe_round_price(best * 0.90, tick_size)
    order = await async_rc.create_order(
        order_type="LIMIT",
        product_id=pid,
        side=0,
        price=price,
        quantity=0.003,
        time_in_force="GTD",
        post_only=False,
        dry_run=True,
        sender=async_rc.chain.address,
        subaccount=subaccounts[0].name,
    )
    assert isinstance(order, async_rc._models.DryRunOrderCreatedDto)


@pytest.mark.asyncio
async def test_rest_market_order_dry(async_rc, network):
    subaccounts = await async_rc.subaccounts()
    pid = (await async_rc.list_products())[0].id
    order = await async_rc.create_order(
        order_type="MARKET",
        product_id=pid,
        side=0,
        quantity=0.003,
        dry_run=True,
        sender=async_rc.chain.address,
        subaccount=subaccounts[0].name,
    )
    assert isinstance(order, async_rc._models.DryRunOrderCreatedDto)


@pytest.mark.asyncio
async def test_rest_market_order_submit(async_rc, network):
    subaccounts = await async_rc.subaccounts()
    pid = (await async_rc.list_products())[0].id
    order = await async_rc.create_order(
        order_type="MARKET",
        product_id=pid,
        side=0,
        quantity=0.003,
        sender=async_rc.chain.address,
        subaccount=subaccounts[0].name,
    )
    assert isinstance(order, async_rc._models.SubmitOrderCreatedDto)


@pytest.mark.asyncio
async def test_rest_market_order_submit_close(async_rc, network):
    subaccounts = await async_rc.subaccounts()
    pid = (await async_rc.list_products())[0].id
    o1 = await async_rc.create_order(
        order_type="MARKET",
        product_id=pid,
        side=0,
        quantity=0.003,
        sender=async_rc.chain.address,
        subaccount=subaccounts[0].name,
    )
    assert isinstance(o1, async_rc._models.SubmitOrderCreatedDto)
    o2 = await async_rc.create_order(
        order_type="MARKET",
        product_id=pid,
        side=1,
        quantity=0,
        reduce_only=True,
        close=True,
        sender=async_rc.chain.address,
        subaccount=subaccounts[0].name,
    )
    assert isinstance(o2, async_rc._models.SubmitOrderCreatedDto)


@pytest.mark.asyncio
async def test_rest_prepare_and_sign_limit_order(async_rc, network):
    subaccounts = await async_rc.list_subaccounts(sender=async_rc.chain.address)
    sub = subaccounts[0]
    products = await async_rc.list_products()
    pid = products[0].id
    onchain_id = products[0].onchain_id
    tick_size = float((await async_rc.products_by_id())[pid].tick_size)
    prices = (await async_rc.list_market_prices(product_ids=[pid]))[0]
    best = float(prices.best_bid_price)
    price = safe_round_price(best * 0.90, tick_size)
    order = await async_rc.prepare_order(
        sender=async_rc.chain.address,
        subaccount=sub.name,
        order_type="LIMIT",
        onchain_id=onchain_id,
        side=0,
        price=price,
        quantity=0.003,
        time_in_force="GTD",
        post_only=False,
        include_signature=True,
    )
    assert isinstance(order, async_rc._models.SubmitOrderDto) and order.signature != ""


@pytest.mark.asyncio
async def test_rest_prepare_and_sign_market_order(async_rc, network):
    subaccounts = await async_rc.list_subaccounts(sender=async_rc.chain.address)
    sub = subaccounts[0]
    products = await async_rc.list_products()
    onchain_id = products[0].onchain_id
    order = await async_rc.prepare_order(
        sender=async_rc.chain.address,
        subaccount=sub.name,
        order_type="MARKET",
        onchain_id=onchain_id,
        side=0,
        price="0",
        quantity=0.003,
        include_signature=True,
    )
    assert isinstance(order, async_rc._models.SubmitOrderDto) and order.signature != ""



from typing import Union, Dict, Any, Optional, List, Callable
import asyncio
import threading
import socketio

from ethereal.models.config import WSBaseConfig
from ethereal.base_client import BaseClient

try:
    import uvloop  # type: ignore

    uvloop_enabled = True
except Exception:
    uvloop_enabled = False


def _create_event_loop() -> asyncio.AbstractEventLoop:
    if uvloop_enabled:
        try:
            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
        except Exception:
            pass
    return asyncio.new_event_loop()


class WSBase(BaseClient):
    """Websocket client base class.

    Args:
        config (Union[Dict[str, Any], WSConfig]): Configuration dictionary or WSConfig object.
            Required fields include:
            - base_url (str): Base URL for websocket requests
            Optional fields include:
            - verbose (bool): Enables debug logging, defaults to False
    """

    def __init__(self, config: Union[Dict[str, Any], WSBaseConfig]):
        super().__init__(config)
        self.config = WSBaseConfig.model_validate(config)
        self.base_url = f"{self.config.base_url}"

        # Initialize socketio client
        self.sio = socketio.AsyncClient(
            logger=self.config.verbose, engineio_logger=self.config.verbose
        )
        self.loop: Optional[asyncio.AbstractEventLoop] = None
        self.thread: Optional[threading.Thread] = None
        self.callbacks: dict[str, List[Callable]] = {}

        # Setup basic event handlers
        @self.sio.event
        async def connect():
            self.logger.info("Connected to server")

        @self.sio.event
        async def disconnect():
            self.logger.info("Disconnected from server")

        # Setup generic message handler
        @self.sio.on(event="*", namespace="*")
        async def catch_all(event, namespace, data):
            self.logger.debug(
                f"Received event: {event} in namespace {namespace} with data: {data}"
            )
            if event in self.callbacks:
                for callback in self.callbacks[event]:
                    try:
                        callback(data)
                    except Exception as e:
                        self.logger.error(f"Error in callback for event {event}: {e}")

    def open(self, namespaces: Optional[List[str]] = None) -> None:
        """Open the websocket client connection."""
        if not self.loop or self.loop.is_closed():
            self.loop = _create_event_loop()
            self.thread = threading.Thread(target=self.loop.run_forever)
            self.thread.daemon = True
            self.thread.start()

        future = asyncio.run_coroutine_threadsafe(
            self.sio.connect(
                self.base_url,
                transports=["websocket"],
                namespaces=namespaces,
            ),
            self.loop,
        )
        future.result()

    def close(self) -> None:
        """Close the websocket client connection."""
        if self.thread and self.loop and not self.loop.is_closed():
            # Disconnect the socket
            future = asyncio.run_coroutine_threadsafe(self.sio.disconnect(), self.loop)
            future.result()

            # Stop the event loop
            self.loop.call_soon_threadsafe(self.loop.stop)
            self.thread.join()
            self.loop.close()
            self.loop = None
            self.thread = None

    def _emit(
        self, event: str, data: Dict[str, Any], namespace: Optional[str] = None
    ) -> Any:
        """Emit an event to the server.

        Args:
            event (str): Event name
            data (Dict[str, Any]): Event data

        Returns:
            Any: Response from the server
        """
        if not self.loop or self.loop.is_closed():
            raise RuntimeError("WebSocket is not connected")

        future = asyncio.run_coroutine_threadsafe(
            self.sio.emit(event, data, callback=True, namespace=namespace), self.loop
        )
        return future.result()



"""Pure async tests for linking and revoking signers."""

import pytest
from eth_account import Account


@pytest.mark.skip(
    reason="Linked signers are rate limited; running repeatedly may fail."
)
@pytest.mark.asyncio
async def test_link_and_revoke_signer(async_rc, async_subaccount, network):
    sname = async_subaccount.name
    sid_any = async_subaccount.id
    account = Account.create()
    link = await async_rc.prepare_linked_signer(
        sender=async_rc.chain.address,
        signer=account.address,
        subaccount=sname,
        subaccount_id=sid_any,
    )
    link = await async_rc.sign_linked_signer(
        link, private_key=async_rc.chain.private_key, signer_private_key=account.key
    )
    res = await async_rc.link_linked_signer(dto=link)
    assert isinstance(res, async_rc._models.SignerDto) and res.signer == account.address
    revoke = await async_rc.prepare_revoke_linked_signer(
        sender=async_rc.chain.address,
        signer=account.address,
        subaccount=sname,
        subaccount_id=sid_any,
    )
    revoke = await async_rc.sign_revoke_linked_signer(revoke)
    rev_res = await async_rc.revoke_linked_signer(dto=revoke)
    assert isinstance(rev_res, async_rc._models.RevokeLinkedSignerDto)


@pytest.mark.asyncio
async def test_prepare_linked_signer(async_rc_ro, async_subaccount, network):
    sname = async_subaccount.name
    sid_any = async_subaccount.id
    account = Account.create()
    link = await async_rc_ro.prepare_linked_signer(
        sender=async_rc_ro.chain.address,
        signer=account.address,
        subaccount=sname,
        subaccount_id=sid_any,
    )
    assert isinstance(link, async_rc_ro._models.LinkSignerDto)
    typed = async_rc_ro._models.LinkSignerDtoData.model_validate(
        link.data.model_dump(by_alias=True)
    )
    assert link.signature == "" and link.signer_signature == ""
    assert (
        typed.sender == async_rc_ro.chain.address
        and typed.signer == account.address
        and typed.subaccount_id == sid_any
    )


@pytest.mark.asyncio
async def test_prepare_and_sign_linked_signer_sender(
    async_rc, async_subaccount, network
):
    sname = async_subaccount.name
    sid_any = async_subaccount.id
    account = Account.create()
    link = await async_rc.prepare_linked_signer(
        sender=async_rc.chain.address,
        signer=account.address,
        subaccount=sname,
        subaccount_id=sid_any,
    )
    link = await async_rc.sign_linked_signer(
        link, private_key=async_rc.chain.private_key
    )
    assert (
        isinstance(link, async_rc._models.LinkSignerDto)
        and link.signature != ""
        and link.signer_signature == ""
    )


@pytest.mark.asyncio
async def test_prepare_and_sign_linked_signer_signer(
    async_rc_ro, async_subaccount, network
):
    sname = async_subaccount.name
    sid_any = async_subaccount.id
    account = Account.create()
    link = await async_rc_ro.prepare_linked_signer(
        sender=async_rc_ro.chain.address,
        signer=account.address,
        subaccount=sname,
        subaccount_id=sid_any,
    )
    link = await async_rc_ro.sign_linked_signer(link, signer_private_key=account.key)
    assert (
        isinstance(link, async_rc_ro._models.LinkSignerDto)
        and link.signature == ""
        and link.signer_signature != ""
    )


@pytest.mark.asyncio
async def test_prepare_and_sign_linked_signer_both(async_rc, async_subaccount, network):
    sname = async_subaccount.name
    sid_any = async_subaccount.id
    account = Account.create()
    link = await async_rc.prepare_linked_signer(
        sender=async_rc.chain.address,
        signer=account.address,
        subaccount=sname,
        subaccount_id=sid_any,
    )
    link = await async_rc.sign_linked_signer(
        link, private_key=async_rc.chain.private_key
    )
    link = await async_rc.sign_linked_signer(link, signer_private_key=account.key)
    assert (
        isinstance(link, async_rc._models.LinkSignerDto)
        and link.signature != ""
        and link.signer_signature != ""
        and link.signature != link.signer_signature
    )


@pytest.mark.asyncio
async def test_prepare_revoke_linked_signer(async_rc_ro, async_subaccount, network):
    sname = async_subaccount.name
    sid_any = async_subaccount.id
    account = Account.create()
    revoke = await async_rc_ro.prepare_revoke_linked_signer(
        sender=async_rc_ro.chain.address,
        signer=account.address,
        subaccount=sname,
        subaccount_id=sid_any,
    )
    assert isinstance(revoke, async_rc_ro._models.RevokeLinkedSignerDto)
    assert (
        revoke.data.sender == async_rc_ro.chain.address
        and revoke.data.signer == account.address
        and revoke.data.subaccount_id == sid_any
        and revoke.signature == ""
    )


@pytest.mark.asyncio
async def test_prepare_and_sign_revoke_linked_signer(
    async_rc, async_subaccount, network
):
    sname = async_subaccount.name
    sid_any = async_subaccount.id
    account = Account.create()
    revoke = await async_rc.prepare_revoke_linked_signer(
        sender=async_rc.chain.address,
        signer=account.address,
        subaccount=sname,
        subaccount_id=sid_any,
    )
    revoke = await async_rc.sign_revoke_linked_signer(revoke)
    assert (
        isinstance(revoke, async_rc._models.RevokeLinkedSignerDto)
        and revoke.signature != ""
    )



import importlib
import inspect
from pydantic import BaseModel
import re

MODELS_MODULE = "ethereal.models.rest"
OUTPUT_FILE = "all_model_docs.txt"


def camel_to_snake(name: str) -> str:
    """
    Convert CamelCase or camelCase string to snake_case.
    Examples:
        >>> camel_to_snake('CamelCase')
        'camel_case'
        >>> camel_to_snake('camelCase')
        'camel_case'
    """
    # First, put an underscore between a lowercase-to-uppercase transition
    s1 = re.sub(r"(.)([A-Z][a-z]+)", r"\1_\2", name)
    # Next, handle transitions like lowercase-or-digit to uppercase
    s2 = re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", s1)
    return s2.lower()


def model_to_doc(model: type[BaseModel]) -> str:
    """
    Render a single-model documentation block in “Args / Other Parameters” style.
    """
    schema = model.schema()
    title = model.__name__
    required_fields = set(schema.get("required", []))
    props = schema.get("properties", {})

    # split into required vs optional
    req_props = {k: v for k, v in props.items() if k in required_fields}
    opt_props = {k: v for k, v in props.items() if k not in required_fields}

    lines = [
        title,
        "=" * len(title),
        "",
        "Args:",
    ]
    for name, meta in req_props.items():
        typ = meta.get("type", "Any")
        desc = meta.get("description", "").rstrip(".")
        lines.append(f"    {camel_to_snake(name)} ({typ}): {desc}. Required.")
    lines += [
        "",
        "Other Parameters:",
    ]
    if opt_props:
        for name, meta in opt_props.items():
            typ = meta.get("type", "Any")
            desc = meta.get("description", "").rstrip(".")
            lines.append(
                f"    {camel_to_snake(name)} ({typ}, optional): {desc}. Optional."
            )
    lines.append(
        "    **kwargs: Additional request parameters accepted by the API. Optional."
    )
    lines.append("\n")
    return "\n".join(lines)


def main():
    mod = importlib.import_module(MODELS_MODULE)

    docs = []
    for name, obj in vars(mod).items():
        if inspect.isclass(obj) and issubclass(obj, BaseModel) and obj is not BaseModel:
            docs.append(model_to_doc(obj))

    with open(OUTPUT_FILE, "w") as f:
        f.write("\n".join(docs))

    print(f"Wrote docs for {len(docs)} models to {OUTPUT_FILE}")


if __name__ == "__main__":
    main()



from typing import List
from ethereal.constants import API_PREFIX
from ethereal.models.rest import FundingDto, ProjectedFundingDto


async def list_funding(self, **kwargs) -> List[FundingDto]:
    """Lists historical funding rates for a product over a specified time range.

    Args:
        product_id (str): Id representing the registered product. Required.
        range (str): Time window to query. One of 'DAY', 'WEEK', or 'MONTH'. Required.

    Other Parameters:
        order (str, optional): Sort order, 'asc' or 'desc'. Optional.
        limit (float, optional): Maximum number of results to return. Optional.
        cursor (str, optional): Pagination cursor for fetching the next page. Optional.
        order_by (str, optional): Field to order by, e.g., 'createdAt'. Optional.
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        List[FundingDto]: Funding rate history objects for the product.
    """
    res = await self.get_validated(
        url_path=f"{API_PREFIX}/funding",
        request_model=self._models.V1FundingGetParametersQuery,
        response_model=self._models.PageOfFundingDtos,
        **kwargs,
    )
    data = [
        self._models.FundingDto(**model.model_dump(by_alias=True)) for model in res.data
    ]
    return data


async def get_projected_funding(self, **kwargs) -> ProjectedFundingDto:
    """Gets the projected funding rate for the next period.

    Args:
        product_id (str): Id representing the registered product. Required.

    Other Parameters:
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        ProjectedFundingDto: Projected funding information for the product.
    """
    res = await self.get_validated(
        url_path=f"{API_PREFIX}/funding/projected",
        request_model=self._models.V1FundingProjectedGetParametersQuery,
        response_model=self._models.ProjectedFundingDto,
        **kwargs,
    )
    return res



import os
import json

from typing import Optional, Dict, Any, Union, List
from pydantic import BaseModel
from ethereal.base_client import BaseClient
from ethereal.models.config import ChainConfig
from ethereal.models.rest import RpcConfigDto, TokenDto
from ethereal.rest.util import ensure_bytes32_hex
from web3 import Web3
from web3.exceptions import Web3Exception, ContractCustomError
from web3.types import TxParams, Nonce
from eth_account import Account
from eth_account.messages import encode_typed_data
from eth_utils import encode_hex
from eth_utils.abi import function_abi_to_4byte_selector


def read_contract(
    chain_id: int, contract_name: str, common: bool = False
) -> Dict[str, Any]:
    if common:
        contract_dir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "contracts",
            "common",
            f"{contract_name}.json",
        )
    else:
        contract_dir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "contracts",
            f"{chain_id}",
            f"{contract_name}.json",
        )
    with open(contract_dir) as f:
        return json.load(f)


class ChainClient(BaseClient):
    """Client for interacting with the blockchain using Web3 functionality.

    Args:
        config (Union[Dict[str, Any], ChainConfig]): Chain configuration
        rpc_config (RpcConfigDto, optional): RPC configuration. Defaults to None.

    Raises:
        Exception: If RPC URL or private key is not specified in the configuration
    """

    def __init__(
        self,
        config: Union[Dict[str, Any], ChainConfig],
        rpc_config: Optional[RpcConfigDto] = None,
        tokens: Optional[List[TokenDto]] = None,
    ):
        super().__init__(config)
        self.config = ChainConfig.model_validate(config)
        self.provider = self._setup_provider()
        self.account = self._setup_account()
        if self.account:
            self.address = self.account.address
            self.private_key = self.config.private_key
        else:
            self.address = self.config.address
            self.private_key = None

        self.chain_id = self.provider.eth.chain_id

        if tokens is not None:
            usde_token = next((t for t in tokens if t.name == "USD"), None)
            if usde_token is None:
                self.logger.warning("USD token not found in the provided tokens list.")
            else:
                usde_address = self.provider.to_checksum_address(usde_token.address)
                usde_abi = (
                    read_contract(self.chain_id, "WUSDe", common=True)
                    if usde_token.erc20_name and "Wrapped" in usde_token.erc20_name
                    else read_contract(self.chain_id, "ERC20", common=True)
                )

                self.usde = self.provider.eth.contract(
                    address=usde_address,
                    abi=usde_abi,
                )
        self.rpc_config = rpc_config

    @property
    def exchange_contract(self):
        return self.provider.eth.contract(
            address=self.rpc_config.domain.verifying_contract,
            abi=read_contract(self.chain_id, "ExchangeGateway"),
        )

    def _setup_provider(self):
        """Set up the Web3 provider.

        Returns:
            Web3: The Web3 provider instance

        Raises:
            Exception: If RPC URL is not specified in the configuration
        """
        # TODO: Support other provider types (e.g. WebSocket)
        if self.config.rpc_url is None:
            raise Exception("RPC URL must be specified in the configuration")
        return Web3(Web3.HTTPProvider(self.config.rpc_url))

    def _setup_account(self):
        """Set up the account.

        Returns:
            Account: The Web3 account instance

        Raises:
            Exception: If private key is not specified in the configuration
        """
        if self.config.private_key is None:
            self.logger.debug("Private key not specified in the configuration")
            return None

        account = self.provider.eth.account.from_key(self.config.private_key)
        if not account:
            raise Exception("Failed to create account from private key")
        if self.config.address and account.address != self.config.address:
            raise Exception(
                f"Private key does not match address specified in the config: {self.config.address}"
            )
        return account

    def _get_tx(self, value=0, to=None) -> TxParams:
        """Get default transaction parameters.

        Args:
            value (int, optional): The value to send. Defaults to 0.
            to (str, optional): The recipient address. Defaults to None.

        Returns:
            TxParams: The transaction parameters
        """
        params: TxParams = {
            "from": self.address,
            "chainId": self.chain_id,
            "value": value,
            "nonce": Nonce(self.get_nonce(self.address)),
        }
        if to is not None:
            params["to"] = to
        return params

    def _decode_error(self, error: ContractCustomError) -> str:
        abi_errors = {
            self.provider.to_hex(function_abi_to_4byte_selector(f)): f.get("name")
            for f in self.exchange_contract.abi
            if f.get("type") == "error"
        }
        data = error.data
        error_signature = data
        return abi_errors.get(error_signature, "Unknown error")

    def _get_by_alias(self, model: BaseModel, alias: str):
        """
        Get a field value by its alias from a Pydantic model.

        Args:
            model (BaseModel): The Pydantic model instance to extract the field from.
            alias (str): The alias of the field to retrieve.

        Returns:
            Any: The value of the field with the specified alias.

        Raises:
            KeyError: If the alias is not found in the model.
        """
        for field_name, field in model.__class__.model_fields.items():
            if field.alias == alias:
                return getattr(model, field_name)
        raise KeyError(f"Alias '{alias}' not found in model {model.__class__.__name__}")

    def get_signature_types(self, rpc_config: RpcConfigDto, primary_type: str):
        """Gets EIP-712 signature types.

        Args:
            rpc_config (RpcConfigDto): RPC configuration.
            primary_type (str): Primary type for the signature.

        Returns:
            dict: Dictionary containing signature type definitions.
        """
        return {
            "EIP712Domain": [
                {"name": "name", "type": "string"},
                {"name": "version", "type": "string"},
                {"name": "chainId", "type": "uint256"},
                {"name": "verifyingContract", "type": "address"},
            ],
            primary_type: self.convert_types(
                self._get_by_alias(rpc_config.signature_types, primary_type),
            ),
        }

    def convert_types(self, type_string: str) -> List[Dict[str, str]]:
        """Converts type string into EIP-712 field format.

        Args:
            type_string (str): String containing type definitions.

        Returns:
            List[Dict[str, str]]: List of field definitions.
        """
        fields = [comp.strip() for comp in type_string.split(",")]
        type_fields = []
        for field in fields:
            field_type, field_name = field.rsplit(" ", 1)
            type_fields.append({"name": field_name, "type": field_type})
        return type_fields

    def add_gas_fees(self, tx: TxParams) -> TxParams:
        """Add gas fee parameters to a transaction.

        Args:
            tx (TxParams): The transaction parameters

        Returns:
            TxParams: The transaction parameters with gas fee parameters added
        """
        if "maxFeePerGas" in tx and "maxPriorityFeePerGas" in tx:
            return tx
        try:
            gas_price = self.provider.eth.gas_price
            max_priority_fee = self.provider.eth.max_priority_fee
            tx["maxFeePerGas"] = gas_price
            tx["maxPriorityFeePerGas"] = max_priority_fee
            return tx
        except Web3Exception as e:
            self.logger.error(f"Failed to add gas: {e}")
            return tx

    def add_gas_limit(self, tx: TxParams) -> TxParams:
        """Add gas limit to a transaction.

        Args:
            tx (TxParams): The transaction parameters

        Returns:
            TxParams: The transaction parameters with gas limit added
        """
        if "gas" in tx:
            return tx
        try:
            gas = self.provider.eth.estimate_gas(tx)
            tx["gas"] = gas
            return tx
        except Web3Exception as e:
            self.logger.error(
                f"Failed to add gas limit: {self._decode_error(e) if isinstance(e, ContractCustomError) else e}"
            )
            raise e

    def submit_tx(self, tx: TxParams) -> str:
        """Submit a transaction.

        Args:
            tx (TxParams): The transaction parameters

        Returns:
            str: The transaction hash
        """
        tx = self.add_gas_fees(tx)
        tx = self.add_gas_limit(tx)
        try:
            signed_tx = self.provider.eth.account.sign_transaction(
                tx, private_key=self.private_key
            )
            tx_hash = self.provider.eth.send_raw_transaction(signed_tx.raw_transaction)
            return encode_hex(tx_hash)
        except Web3Exception as e:
            self.logger.error(f"Failed to submit transaction: {e}")
            raise e

    def get_nonce(self, address: str) -> int:
        """Get the nonce for a given address.

        Args:
            address (str): The address to get the nonce for

        Returns:
            int: The nonce, or -1 if failed
        """
        try:
            return self.provider.eth.get_transaction_count(address)
        except Web3Exception as e:
            self.logger.error(f"Failed to get nonce: {e}")
            return -1

    def get_balance(self, address: str) -> int:
        """Get the balance for a given address.

        Args:
            address (str): The address to get the balance for

        Returns:
            int: The balance, or -1 if failed
        """
        try:
            return self.provider.eth.get_balance(address)
        except Web3Exception as e:
            self.logger.error(f"Failed to get balance: {e}")
            return -1

    def get_token_balance(self, address: str, token_address: str) -> int:
        """Get the token balance for a given address.

        Args:
            address (str): The address to get the token balance for
            token_address (str): The token address

        Returns:
            int: The token balance, or -1 if failed
        """
        try:
            contract = self.provider.eth.contract(
                address=token_address,
                abi=read_contract(self.chain_id, "ERC20", common=True),
            )
            return contract.functions.balanceOf(address).call()
        except Web3Exception as e:
            self.logger.error(f"Failed to get token balance: {e}")
            return -1

    def sign_message(
        self,
        private_key: str,
        domain: dict,
        types: dict,
        primary_type: str,
        message: dict,
    ):
        """Sign an EIP-712 typed data message.

        Args:
            private_key (str): private key to sign the message with
            domain (dict): domain parameters including name, version, chainId, and verifyingContract
            types (dict): type definitions for the structured data
            primary_type (str): primary type for the signature
            message (dict): message data to be signed

        Returns:
            str: the hexadecimal signature string prefixed with '0x'
        """
        # A type fix for the domain
        domain["chainId"] = int(domain["chainId"])

        # Preparing the full message as per EIP-712
        full_message = {
            "types": types,
            "primaryType": primary_type,
            "domain": domain,
            "message": message,
        }

        encoded_message = encode_typed_data(full_message=full_message)

        # Signing the message
        signed_message = Account.sign_message(encoded_message, private_key)
        return "0x" + signed_message.signature.hex()

    def deposit_usde(
        self,
        amount: float,
        address: Optional[str] = None,
        submit: bool = False,
        account_name: Optional[str] = None,
        account_name_bytes: Optional[str] = None,
    ) -> Union[TxParams, str]:
        """Submit a deposit transaction.

        Args:
            amount (float): The amount to deposit
            address (str, optional): The address to deposit to. Defaults to None.
            submit (bool, optional): Whether to submit the transaction. Defaults to False.
            account_name (str, optional): The account name. Defaults to "primary".
            account_name_bytes (str, optional): The account name as a hex string (bytes32). Defaults to None.

        Returns:
            Union[TxParams, str]: The transaction parameters or transaction hash if submit=True

        Raises:
            ValueError: If both account_name and account_name_bytes are provided
        """
        if address is None:
            address = self.address

        if account_name is None and account_name_bytes is None:
            account_name = "primary"

        # Validate inputs
        if account_name is not None and account_name_bytes is not None:
            raise ValueError("Cannot provide both account_name and account_name_bytes")

        try:
            # params
            if account_name_bytes is not None:
                subaccount = ensure_bytes32_hex(account_name_bytes)
            else:
                subaccount = self.provider.to_hex(text=account_name).ljust(66, "0")
            amount = self.provider.to_wei(amount, "ether")
            referral_code = self.provider.to_hex(0).ljust(66, "0")

            # prepare the tx
            tx = self._get_tx(to=self.exchange_contract.address, value=amount)
            tx["data"] = self.exchange_contract.encode_abi(
                "depositUsd", args=[subaccount, referral_code]
            )

            if submit:
                return self.submit_tx(tx)
            else:
                return tx

        except Web3Exception as e:
            self.logger.error(
                f"Failed to prepare deposit transaction: {self._decode_error(e) if isinstance(e, ContractCustomError) else e}"
            )
            raise e

    def finalize_withdraw(
        self,
        address: Optional[str] = None,
        submit: Optional[bool] = False,
        account_name: Optional[str] = "primary",
    ) -> Union[TxParams, str]:
        """Finalize a withdrawal.

        Args:
            address (str, optional): The address to deposit to. Defaults to None.
            submit (bool, optional): Whether to submit the transaction. Defaults to False.
            account_name (str, optional): The name of the account. Defaults to "primary".

        Returns:
            Union[TxParams, str]: The transaction parameters or transaction hash if submit=True
        """
        if address is None:
            address = self.address
        try:
            # params
            subaccount = self.provider.to_hex(text=account_name).ljust(66, "0")

            # prepare the tx
            tx = self._get_tx(to=self.exchange_contract.address)
            tx["data"] = self.exchange_contract.encode_abi(
                "finalizeWithdraw", args=[address, subaccount]
            )

            if submit:
                return self.submit_tx(tx)
            else:
                return tx

        except Web3Exception as e:
            self.logger.error(
                f"Failed to prepare finalizeWithdraw transaction: {self._decode_error(e) if isinstance(e, ContractCustomError) else e}"
            )
            raise e



from typing import List, Optional
from uuid import UUID
from ethereal.constants import API_PREFIX
from ethereal.rest.util import generate_nonce
from ethereal.models.rest import (
    SignerDto,
    AccountSignerQuotaDto,
    LinkSignerDto,
    RevokeLinkedSignerDto,
)
import time


async def list_signers(
    self,
    **kwargs,
) -> List[SignerDto]:
    """Lists linked signers for a subaccount.

    Args:
        subaccount_id (str): UUID of the subaccount. Required.

    Other Parameters:
        order (str, optional): Sort order, 'asc' or 'desc'. Optional.
        limit (float, optional): Maximum number of results to return. Optional.
        cursor (str, optional): Pagination cursor for fetching the next page. Optional.
        order_by (str, optional): Field to order by (e.g., 'createdAt'). Optional.
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        List[SignerDto]: Linked signer records.
    """
    res = await self.get_validated(
        url_path=f"{API_PREFIX}/linked-signer",
        request_model=self._models.V1LinkedSignerGetParametersQuery,
        response_model=self._models.PageOfSignersDto,
        **kwargs,
    )
    data = [
        self._models.SignerDto(**model.model_dump(by_alias=True)) for model in res.data
    ]
    return data


async def get_signer(
    self,
    id: UUID,
    **kwargs,
) -> SignerDto:
    """Gets a specific linked signer by ID.

    Args:
        id (str): UUID of the linked signer. Required.

    Other Parameters:
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        SignerDto: Linked signer details.
    """
    endpoint = f"{API_PREFIX}/linked-signer/{id}"
    res = await self.get(endpoint, **kwargs)
    return self._models.SignerDto.model_validate(res)


async def get_signer_quota(
    self,
    **kwargs,
) -> AccountSignerQuotaDto:
    """Gets the signer quota for a subaccount.

    Args:
        subaccount_id (str): UUID of the subaccount. Required.

    Other Parameters:
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        AccountSignerQuotaDto: Remaining quota information for the subaccount.
    """
    res = await self.get_validated(
        url_path=f"{API_PREFIX}/linked-signer/quota",
        request_model=self._models.V1LinkedSignerQuotaGetParametersQuery,
        response_model=self._models.AccountSignerQuotaDto,
        **kwargs,
    )
    return res


async def prepare_linked_signer(
    self,
    sender: str,
    signer: str,
    subaccount: str,
    subaccount_id: UUID,
    signer_signature: str = "",
    include_signature: bool = False,
    **kwargs,
) -> LinkSignerDto:
    """Prepares the payload for linking a signer, optionally including a signature.

    Args:
        sender (str): Owner address initiating the link. Required.
        signer (str): Address of the signer being linked. Required.
        subaccount (str): Hex-encoded subaccount name. Required.
        subaccount_id (str): UUID of the subaccount. Required.
        signer_signature (str): Signature from the signer address. Optional.
        include_signature (bool): If True, sign with the owner's key as well. Optional.

    Other Parameters:
        nonce (str, optional): Custom nonce for signing. Optional.
        signed_at (int, optional): Seconds since epoch for the signature timestamp. Optional.
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        LinkSignerDto: Prepared (and optionally signed) link payload.
    """
    nonce = kwargs.get("nonce") or generate_nonce()
    signed_at = kwargs.get("signed_at") or int(time.time())
    data = {
        "sender": sender,
        "signer": signer,
        "subaccount": subaccount,
        "subaccountId": subaccount_id,
        "nonce": nonce,
        "signedAt": signed_at,
    }
    data_model = self._models.LinkSignerDtoData.model_validate(data)

    # Prepare dto
    dto_data = {
        "data": data_model.model_dump(mode="json", by_alias=True),
        "signature": "",
        "signerSignature": signer_signature,
    }
    dto = self._models.LinkSignerDto.model_validate(dto_data, by_alias=True)
    if include_signature:
        dto = await self.sign_linked_signer(dto, private_key=self.chain.private_key)
    return dto


async def sign_linked_signer(
    self,
    link_to_sign: LinkSignerDto,
    signer_private_key: Optional[str] = None,
    private_key: Optional[str] = None,
) -> LinkSignerDto:
    """Signs the link-signer payload with the signer and/or owner key.

    Args:
        link_to_sign (LinkSignerDto): Prepared link payload. Required.
        signer_private_key (str, optional): Signer's private key for cosigning. Optional.
        private_key (str, optional): Owner's private key override. Optional.

    Returns:
        LinkSignerDto: DTO with signature fields populated.

    Raises:
        ValueError: If no chain client or private key is available.
    """
    if not hasattr(self, "chain") or not self.chain:
        raise ValueError("No chain client available for signing")
    if not private_key and not self.chain.private_key and not signer_private_key:
        raise ValueError("No private key available for signing")
    elif not private_key:
        private_key = self.chain.private_key

    # Prepare message for signing
    message = link_to_sign.data.model_dump(mode="json", by_alias=True)
    message["signedAt"] = int(message["signedAt"])

    primary_type = "LinkSigner"
    domain = self.rpc_config.domain.model_dump(mode="json", by_alias=True)
    types = self.chain.get_signature_types(self.rpc_config, primary_type)

    if signer_private_key:
        signer_signature = self.chain.sign_message(
            signer_private_key, domain, types, primary_type, message
        )
        link_to_sign.signer_signature = signer_signature
    if self.chain.private_key:
        signature = self.chain.sign_message(
            private_key, domain, types, primary_type, message
        )
        link_to_sign.signature = signature
    return link_to_sign


async def link_linked_signer(
    self,
    dto: LinkSignerDto,
    **kwargs,
) -> SignerDto:
    """Submits a prepared and signed link-signer payload.

    Args:
        dto (LinkSignerDto): Prepared and signed link payload. Required.

    Other Parameters:
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        SignerDto: Linked signer record after submission.
    """
    endpoint = f"{API_PREFIX}/linked-signer/link"
    res = await self.post(
        endpoint,
        data=dto.model_dump(mode="json", by_alias=True, exclude_none=True),
        **kwargs,
    )
    return self._models.SignerDto.model_validate(res)


async def prepare_revoke_linked_signer(
    self,
    sender: str,
    signer: str,
    subaccount: str,
    subaccount_id: UUID,
    include_signature: bool = False,
    **kwargs,
) -> RevokeLinkedSignerDto:
    """Prepares the payload for revoking a linked signer, optionally signing it.

    Args:
        sender (str): Owner address initiating the revoke. Required.
        signer (str): Signer address being revoked. Required.
        subaccount (str): Hex-encoded subaccount name. Required.
        subaccount_id (str): UUID of the subaccount. Required.
        include_signature (bool): If True, sign with the owner's key. Optional.

    Other Parameters:
        nonce (str, optional): Custom nonce for signing. Optional.
        signed_at (int, optional): Seconds since epoch for the signature timestamp. Optional.
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        RevokeLinkedSignerDto: Prepared (and optionally signed) revoke payload.
    """
    nonce = kwargs.get("nonce") or generate_nonce()
    signed_at = kwargs.get("signed_at") or int(time.time())
    data = {
        "sender": sender,
        "signer": signer,
        "subaccount": subaccount,
        "subaccountId": subaccount_id,
        "nonce": nonce,
        "signedAt": signed_at,
    }
    data_model = self._models.RevokeLinkedSignerDtoData.model_validate(data)
    dto = self._models.RevokeLinkedSignerDto.model_validate(
        {"data": data_model.model_dump(mode="json", by_alias=True), "signature": ""}
    )
    if include_signature:
        dto = await self.sign_revoke_linked_signer(dto)
    return dto


async def sign_revoke_linked_signer(
    self,
    revoke_to_sign: RevokeLinkedSignerDto,
) -> RevokeLinkedSignerDto:
    """Signs the revoke-linked-signer payload with the owner's key.

    Args:
        revoke_to_sign (RevokeLinkedSignerDto): Prepared revoke payload. Required.

    Returns:
        RevokeLinkedSignerDto: DTO with signature populated.

    Raises:
        ValueError: If no chain client or private key is available.
    """
    if not hasattr(self, "chain") or not self.chain:
        raise ValueError("No chain client available for signing")
    if not self.chain.private_key:
        raise ValueError("No private key available for signing")

    # Prepare message for signing
    message = revoke_to_sign.data.model_dump(mode="json", by_alias=True)
    message["signedAt"] = int(message["signedAt"])

    primary_type = "RevokeLinkedSigner"
    domain = self.rpc_config.domain.model_dump(mode="json", by_alias=True)
    types = self.chain.get_signature_types(self.rpc_config, primary_type)
    revoke_to_sign.signature = self.chain.sign_message(
        self.chain.private_key, domain, types, primary_type, message
    )
    return revoke_to_sign


async def revoke_linked_signer(
    self,
    dto: RevokeLinkedSignerDto,
    **kwargs,
) -> SignerDto:
    """Submits a prepared and signed revoke-linked-signer payload.

    Args:
        dto (RevokeLinkedSignerDto): Prepared and signed revoke payload. Required.

    Other Parameters:
        **kwargs: Additional request parameters accepted by the API. Optional.

    Returns:
        SignerDto: Signer record reflecting revocation.
    """
    endpoint = f"{API_PREFIX}/linked-signer/revoke"
    res = await self.post(
        endpoint,
        data=dto.model_dump(mode="json", by_alias=True, exclude_none=True),
        **kwargs,
    )
    return self._models.SignerDto.model_validate(res)



import os
import sys
import json
import copy
import argparse
import requests
import subprocess
from pathlib import Path
from typing import Any, Dict, List

# Network configuration mapping
NETWORK_CONFIGS = {
    "mainnet": {
        "openapi_urls": {
            "primary": "https://api.ethereal.trade/openapi.json",
            "archive": "https://archive.ethereal.trade/openapi.json",
        },
        "output_dir": "ethereal/models/testnet",
        "output_file": "rest.py",
    },
    "testnet": {
        "openapi_urls": {
            "primary": "https://api.etherealtest.net/openapi.json",
            "archive": "https://archive.etherealtest.net/openapi.json",
        },
        "output_dir": "ethereal/models/testnet",
        "output_file": "rest.py",
    },
    "devnet": {
        "openapi_urls": {
            "primary": "https://api.etherealdev.net/openapi.json",
            "archive": "https://archive.etherealdev.net/openapi.json",
        },
        "output_dir": "ethereal/models/devnet",
        "output_file": "rest.py",
    },
}


def parse_args():
    parser = argparse.ArgumentParser(
        description="Generate Pydantic types from OpenAPI spec"
    )
    parser.add_argument(
        "--network",
        type=str,
        choices=list(NETWORK_CONFIGS.keys()),
        default="testnet",
        help="Network to generate types for (default: testnet)",
    )
    parser.add_argument(
        "--url", type=str, help="Custom OpenAPI spec URL (overrides network default)"
    )
    return parser.parse_args()


def get_config(args) -> Dict[str, Any]:
    """Get configuration based on arguments"""
    config = copy.deepcopy(NETWORK_CONFIGS[args.network])

    if args.url:
        config["openapi_urls"] = {"custom": args.url}

    # Generate network-specific file names
    config["output_dir"] = f"ethereal/models/{args.network}"
    config["output_file"] = "rest.py"

    return config


def _merge_dict(target: Dict[str, Any], source: Dict[str, Any], context: str) -> None:
    """Merge source dict into target dict, raising if duplicate keys differ."""
    for key, value in source.items():
        if key in target:
            if target[key] != value:
                raise ValueError(f"Conflict merging OpenAPI specs at {context}.{key}")
        else:
            target[key] = copy.deepcopy(value)


def merge_openapi_specs(specs: List[Dict[str, Any]]) -> Dict[str, Any]:
    """Combine multiple OpenAPI specs into a single spec."""
    if not specs:
        raise ValueError("No OpenAPI specs provided for merging")

    merged = copy.deepcopy(specs[0])

    for spec in specs[1:]:
        if "paths" in spec:
            merged.setdefault("paths", {})
            _merge_dict(merged["paths"], spec["paths"], "paths")

        if "components" in spec:
            merged.setdefault("components", {})
            for section_name, section_value in spec["components"].items():
                merged["components"].setdefault(section_name, {})
                if not isinstance(section_value, dict):
                    merged["components"][section_name] = copy.deepcopy(section_value)
                else:
                    _merge_dict(
                        merged["components"][section_name],
                        section_value,
                        f"components.{section_name}",
                    )

        if "tags" in spec:
            merged.setdefault("tags", [])
            existing_named_tags = {
                tag.get("name")
                for tag in merged["tags"]
                if isinstance(tag, dict) and "name" in tag
            }
            for tag in spec["tags"]:
                if isinstance(tag, dict) and "name" in tag:
                    if tag["name"] not in existing_named_tags:
                        merged["tags"].append(copy.deepcopy(tag))
                        existing_named_tags.add(tag["name"])
                elif tag not in merged["tags"]:
                    merged["tags"].append(copy.deepcopy(tag))

    return merged


def fetch_openapi_spec(url: str) -> Dict[str, Any]:
    """Fetch and parse an OpenAPI spec from the provided URL."""
    response = requests.get(url)
    response.raise_for_status()

    try:
        return response.json()
    except ValueError as exc:
        raise ValueError(f"Invalid JSON received from {url}") from exc


def prepare_spec(spec: Dict[str, Any], alias: str) -> Dict[str, Any]:
    """Apply alias-specific transformations before merging."""
    prepared = copy.deepcopy(spec)

    if alias != "primary" and "paths" in prepared:
        transformed_paths: Dict[str, Any] = {}
        for path_key, path_value in prepared["paths"].items():
            new_key = (
                f"/{alias}{path_key}"
                if not path_key.startswith(f"/{alias}")
                else path_key
            )
            if (
                new_key in transformed_paths
                and transformed_paths[new_key] != path_value
            ):
                raise ValueError(
                    f"Conflict transforming paths for alias '{alias}' at '{new_key}'"
                )
            transformed_paths[new_key] = path_value
        prepared["paths"] = transformed_paths

    return prepared


def generate_types(network: str, config: Dict[str, Any]):
    """Generate types for a specific network by combining all configured specs."""
    print(f"Generating types for {network}...")

    urls = config.get("openapi_urls")
    if not urls:
        print("No OpenAPI URLs configured; nothing to generate.")
        sys.exit(1)

    specs: List[Dict[str, Any]] = []
    for alias, url in urls.items():
        print(f"Fetching {alias} OpenAPI spec from: {url}")
        try:
            spec = fetch_openapi_spec(url)
        except Exception as exc:
            print(f"Error fetching OpenAPI spec ({alias}): {exc}")
            sys.exit(1)
        try:
            prepared = prepare_spec(spec, alias)
        except ValueError as exc:
            print(f"Error preparing OpenAPI spec ({alias}): {exc}")
            sys.exit(1)
        specs.append(prepared)

    try:
        combined_spec = merge_openapi_specs(specs)
    except ValueError as exc:
        print(f"Error merging OpenAPI specs: {exc}")
        sys.exit(1)

    # Create output directory if it doesn't exist
    output_dir = Path(config["output_dir"])
    output_dir.mkdir(parents=True, exist_ok=True)

    # Write the combined spec to a temporary file
    temp_spec_file = f"openapi_{network}.json"
    with open(temp_spec_file, "w") as f:
        json.dump(combined_spec, f)

    # Construct output path
    output_path = output_dir / config["output_file"]

    # Run datamodel-codegen
    result = subprocess.run(
        [
            "uv",
            "run",
            "datamodel-codegen",
            "--input",
            temp_spec_file,
            "--output",
            str(output_path),
            "--input-file-type",
            "openapi",
            "--openapi-scopes",
            "paths",
            "schemas",
            "parameters",
            "--output-model-type",
            "pydantic_v2.BaseModel",
            "--snake-case-field",
        ],
        capture_output=True,
        text=True,
    )

    if result.returncode != 0:
        print(f"Error generating types: {result.stderr}")
        os.remove(temp_spec_file)
        sys.exit(1)
    else:
        print(f"Generated types successfully at: {output_path}")

    # Replace all instances of '0', or "0", with Decimal("0"),
    with open(output_path, "r") as f:
        content = f.read()
    content = content.replace("'0',", 'Decimal("0"),')
    content = content.replace('"0",', 'Decimal("0"),')
    with open(output_path, "w") as f:
        f.write(content)

    # Remove the temporary openapi.json file
    os.remove(temp_spec_file)


def main():
    args = parse_args()
    config = get_config(args)
    generate_types(args.network, config)


if __name__ == "__main__":
    main()
